#include "isegapi.h"
#include "version.h"
#include "kdatacache.h"
#include "socket_can/mainthread.h"

#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <libsocketcan.h>
#include <locale.h>
#include <ctype.h>

static char m_interface[128];
static int m_interfaceIndex = 0;

#define WRONG_INDEX 999
/**
 * @brief Returns the interfaxe index:
 * @return - index 0, 1, ..
 *         - error 999
 */
static int getInterfaceIndex()
{
	if (
			( (strncmp(m_interface, "can", 3) == 0) && (atoi(&m_interface[4]) < 16) ) ||
			strncmp(m_interface, "usb", 3) == 0 ||
			strncmp(m_interface, "com", 3) == 0
	) {
		return atoi(&m_interface[3]);
	}

	return WRONG_INDEX;
}

/**
 * @brief Returns the library four digit version number as unsigned int value.
 *
 * This allows to check for a specific version number.
 * @since 1.1
 * @return The version number as unsigned int value, e.g. 0x01010000
 */
unsigned int iseg_getVersion(void)
{
	const unsigned int version[] = { VER_FILEVERSION };

	return (version[0] << 24) + (version[1] << 16) + (version[2] << 8) + version[3];
}

/**
 * @brief Returns the library four digit version number as null-terminated string.
 *
 * The digits are separated by dots.
 * @since 1.1
 * @return The version number, e.g. "1.1.0.3"
 */
const char *iseg_getVersionString(void)
{
	return VER_FILEVERSION_STR;
}

/**
 * @brief Returns the state of the connected interface.
 * @param property
 * @return
 */
static char *iseg_getLineProperty(const char *property, uint8_t index)
{
	int state;

	//TODO other interfaces
	if (
			strcmp(property, "Status") != 0 &&
			index == getInterfaceIndex()
	) {
		return STATE_UNDEFINED_STRING;
	}

	/* first we check if we can restart the device at all */
	if ((can_get_state(iseg_getConnectedInterface(), &state)) < 0) {
		return CAN_STATE_SERIOUSLY_ERROR_STRING;
	} else {
		switch (state) {

		case CAN_STATE_ERROR_ACTIVE:
			return CAN_STATE_ERROR_ACTIVE_STRING;

		case CAN_STATE_ERROR_WARNING:
			return CAN_STATE_ERROR_WARNING_STRING;

		case CAN_STATE_ERROR_PASSIVE:
			return CAN_STATE_ERROR_PASSIVE_STRING;

		case CAN_STATE_BUS_OFF:
			return CAN_STATE_BUS_OFF_STRING;

		case CAN_STATE_STOPPED:
			return CAN_STATE_STOPPED_STRING;

		case CAN_STATE_SLEEPING:
			return CAN_STATE_SLEEPING_STRING;

		default:
			return STATE_UNDEFINED_STRING;
		}
	}
}

uint32_t iseg_getBitRate(bool *ok)
{
	struct can_bittiming bt;

	if ((can_get_bittiming(iseg_getConnectedInterface(), &bt)) < 0) {
		*ok = false;
		return 0;
	} else {
		*ok = true;
		return bt.bitrate;
	}
}

/**
 * @brief Connects the specified interface to an iseg high voltage device.
 * @param interface - can0, can1, .. can15; usb0, usb1; com1; ip_address
 * @return - 0 = OK
 *         - 1 = ERROR
 */
int iseg_connect(const char *interface)
{
	strncpy(m_interface, interface, sizeof m_interface);
	m_interfaceIndex = getInterfaceIndex();

	kdatacache_clear();

	setlocale(LC_ALL, "C");

	if (m_interfaceIndex == WRONG_INDEX)
		return ISEG_ERROR;

	if (!ksocketcan_start(m_interface))
		return ISEG_ERROR;

	kdatacache_createSystemTimeStamp(true, DataCacheColumnSystemCanStatus);
	kdatacache_createSystemTimeStamp(true, DataCacheColumnSystemCanBitRate);
	return ISEG_OK;
}

int iseg_disconnect(void)
{
	return ksocketcan_stop();
}

char *iseg_getConnectedInterface(void)
{
	return m_interface;
}

enum INDEX_NUMBER {
	LINE_INDEX = 0,
	MODULE_INDEX,
	CHANNEL_INDEX,
	ITEM_INDEX,
	BIT_ARRAY_INDEX,
	BYTE_ARRAY_INDEX,
	INDEX_SIZE
} INDEX_NUMBER;

static bool configureBitArrayIndex(uint16_t *objectAddress, char *p_object)
{
	char *colon = strchr(p_object, ':');
	if (!colon)
		return false;

	objectAddress[BIT_ARRAY_INDEX] = atoi(colon +1);
	*colon = 0;
	return true;
}

static bool configureByteArrayIndex(uint16_t *objectAddress, char *p_object)
{
	char *openingBrace = strchr(p_object, '[');
	if (!openingBrace)
		return false;

	char *closingBrace = strchr(openingBrace, ']');
	if (closingBrace)
		*closingBrace = 0;
	objectAddress[BYTE_ARRAY_INDEX] = atoi(openingBrace + 1);
	*openingBrace = 0;
	return true;
}

enum {
	ObjectSystemParts = 1,
	ObjectLineParts,
	ObjectModuleParts,
	ObjectChannelParts,
	NoneObjectParts
};

uint8_t split(uint16_t *objectAddress, const char *itemObject, char separator)
{
	char object[FULLY_QUALIFIED_OBJECT_SIZE] = { 0 };
	char *p_object;
	uint8_t  separatorNumber = 0;
	uint8_t  pos = 0;
	uint8_t  idx = 0;
	uint16_t  address = 0;

	strncpy(object, itemObject, FULLY_QUALIFIED_OBJECT_SIZE);
	object[FULLY_QUALIFIED_OBJECT_SIZE - 1] = 0;
	p_object = object;
	objectAddress[BIT_ARRAY_INDEX] = NO_BIT_ARRAY_INDEX;
	for (; pos < FULLY_QUALIFIED_OBJECT_SIZE; pos++, idx++) {

		if (p_object[idx] == 0) {
			break;
		}

		if (separatorNumber >= ObjectChannelParts) {
			break;
		}
		if (!strchr(p_object, separator)) {
			if (++separatorNumber == ObjectChannelParts) {
				configureBitArrayIndex(objectAddress, p_object);
				objectAddress[ITEM_INDEX] = kdatacache_getChannelColumn(p_object);
				return separatorNumber;
			}
			if (separatorNumber == ObjectModuleParts) {
				if (!configureBitArrayIndex(objectAddress, p_object))
					configureByteArrayIndex(objectAddress, p_object);
				if (
						objectAddress[MODULE_INDEX] >= CRATE_ADDESS_OFFSET &&
						objectAddress[MODULE_INDEX] < CRATE_ADDESS_OFFSET + DataCacheMaxCrates
				) {
					objectAddress[ITEM_INDEX] = kdatacache_getCrateColumn(p_object);
					return separatorNumber;
				}
				objectAddress[ITEM_INDEX] = kdatacache_getModuleColumn(p_object);
				return separatorNumber;
			}

			if (separatorNumber == ObjectLineParts) {
				configureBitArrayIndex(objectAddress, p_object);
				objectAddress[ITEM_INDEX] = kdatacache_getDeviceColumn(p_object);
				return separatorNumber;
			}

			if (separatorNumber == ObjectSystemParts) {
				if (!configureBitArrayIndex(objectAddress, p_object))
					configureByteArrayIndex(objectAddress, p_object);
				objectAddress[ITEM_INDEX] = kdatacache_getSystemColumn(p_object);
				return separatorNumber;
			}

			return separatorNumber;
		}

		if (p_object[idx] == separator) {
			if (idx == 0) {
				address = -1;		//alle angeschlossenen Datenpunkte
			}

			if (separatorNumber == 0 && address >= DataCacheMaxLines)
				return NoneObjectParts;
			if (
					separatorNumber == 1 &&
					(
						(address >= DataCacheMaxModules && address < CRATE_ADDESS_OFFSET) ||
						(address >= (CRATE_ADDESS_OFFSET + DataCacheMaxCrates))
					)
			)
				return NoneObjectParts;
			if (separatorNumber == 2 && address >= DataCacheMaxChannels)
				return NoneObjectParts;

			objectAddress[separatorNumber] = address;
			if (++separatorNumber >= 4) {
				break;
			}

			p_object = (p_object + idx + 1);

			address = 0;
			idx = -1;
			continue;
		}
		address *= 10;
		address += p_object[idx] - '0';
	}

	return NoneObjectParts;
}

static int setItemValue(const IsegItemValue *item, uint16_t objectAddress[5], u_int16_t number)
{
	if (number == ObjectChannelParts) {
		DataType type = kdatacache_getChannelDataTypeEnum(objectAddress[ITEM_INDEX]);
		if (type == DataTypeR4) {
			kdatacache_setChannelValueR4(atof(item->value), objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
										 objectAddress[CHANNEL_INDEX], objectAddress[ITEM_INDEX]);
			return ISEG_OK;
		}

		kdatacache_setChannelValueUI4(atoi(item->value), objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
										 objectAddress[CHANNEL_INDEX], objectAddress[ITEM_INDEX], objectAddress[BIT_ARRAY_INDEX]);
		return ISEG_OK;
	}

	if (number == ObjectModuleParts) {
		if (objectAddress[MODULE_INDEX] >= CRATE_ADDESS_OFFSET) {
			bool isEch238 = false;
			char quality[QUALITY_SIZE];
			CrateCache *crateCacheValue;

			objectAddress[MODULE_INDEX] -= CRATE_ADDESS_OFFSET;

			strncpy(quality, kdatacache_getCrateQuality(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
															  objectAddress[ITEM_INDEX], isEch238), QUALITY_SIZE);

			isEch238 = (strncmp(quality, ISEG_ITEM_QUALITY_INITIALIZE, QUALITY_SIZE)) ? 0 : 1;
			crateCacheValue = kdatacache_crateCacheValue(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX], isEch238);

			DataType type = kdatacache_getCrateDataTypeEnum(objectAddress[ITEM_INDEX]);
			if (type == DataTypeUI4 || type == DataTypeUI1 || type == DataTypeBool) {
				kdatacache_setCrateValueUI4(atoi(item->value), crateCacheValue, objectAddress[ITEM_INDEX], objectAddress[BIT_ARRAY_INDEX]);
				kdataCache_setCrateCacheStateNewValue(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
													  objectAddress[ITEM_INDEX], isEch238);
				return ISEG_OK;
			}

			return ISEG_ERROR;

		}

		DataType type = kdatacache_getModuleDataTypeEnum(objectAddress[ITEM_INDEX]);
		if (type == DataTypeR4) {
			kdatacache_setModuleValueR4(atof(item->value),
										objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX], objectAddress[ITEM_INDEX]);
			return ISEG_OK;
		}

		if (type == DataTypeString) {
			kdatacache_setModuleValueString(item->value,
											objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX], objectAddress[ITEM_INDEX]);
			return ISEG_OK;
		}

		kdatacache_setModuleValueUI4(atoi(item->value), objectAddress[LINE_INDEX],
									 objectAddress[MODULE_INDEX], objectAddress[ITEM_INDEX], objectAddress[BIT_ARRAY_INDEX]);
		return ISEG_OK;
	}

	if (number == ObjectSystemParts) {
		DataType type = kdatacache_getSystemDataTypeEnum(objectAddress[ITEM_INDEX]);
		if (type == DataTypeUI4) {
			kdatacache_setSystemValue(atoi(item->value), objectAddress[ITEM_INDEX]);
			return ISEG_OK;
		}

		if (type == DataTypeString) {
			kdatacache_setSystemValueString(item->value, objectAddress[ITEM_INDEX]);
			return ISEG_OK;
		}
	}

	return ISEG_ERROR;
}

/**
 * @brief writes an item value, specified by the item object, into the data cache
 * @param item - item struct
 * @return - 0 = ok
 *         - 1 = error
 */
int iseg_setItemValue(const IsegItemValue *item)
{
	int result;
	uint16_t  objectAddress[INDEX_SIZE] = { 0 };

	int number = split(objectAddress, item->object, '.');

	if (objectAddress[LINE_INDEX] < m_interfaceIndex) {
		//printf("Error API %s %s\n", item->object, item->value);
		return ISEG_ERROR;
	}

	objectAddress[LINE_INDEX] -= m_interfaceIndex;

	if (number == NoneObjectParts) {
		//printf("Error API %s %s\n", item->object, item->value);
		return ISEG_ERROR;
	}

	kmutex_lock();
	result = setItemValue(item, objectAddress, number);
	kmutex_unlock();

	//if (result != ISEG_OK) {
	//	printf("Error API %s %s\n", item->object, item->value);
	//}
	return result;
}

static void itemValue(IsegItemValue *item, uint16_t *objectAddress, uint16_t number)
{
	if (number == ObjectChannelParts) {
		if (objectAddress[ITEM_INDEX] >= DataCacheColumnChannelSize) {
			strncpy(item->quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
			return;
		}

		DataType type = kdatacache_getChannelDataTypeEnum(objectAddress[ITEM_INDEX]);
		strncpy(item->timeStampLastRefreshed,
				kdatacache_getChannelTimeStampLastRefreshed(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
															objectAddress[CHANNEL_INDEX], objectAddress[ITEM_INDEX]), TIME_SIZE);
		strncpy(item->timeStampLastChanged,
				kdatacache_getChannelTimeStampLastChanged(objectAddress[LINE_INDEX],
														  objectAddress[MODULE_INDEX], objectAddress[CHANNEL_INDEX],
														  objectAddress[ITEM_INDEX]), TIME_SIZE);
		strncpy(item->quality,
				kdatacache_getChannelQuality(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
											 objectAddress[CHANNEL_INDEX], objectAddress[ITEM_INDEX]), QUALITY_SIZE);

		if (type == DataTypeR4) {
			float value = kdatacache_getChannelValueR4(objectAddress[LINE_INDEX],
													   objectAddress[MODULE_INDEX], objectAddress[CHANNEL_INDEX], objectAddress[ITEM_INDEX]);
			snprintf(item->value, VALUE_SIZE, "%g", value);
			return;
		}

		uint32_t value = kdatacache_getChannelValueUI4(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
													   objectAddress[CHANNEL_INDEX], objectAddress[ITEM_INDEX], objectAddress[BIT_ARRAY_INDEX]);
		snprintf(item->value, VALUE_SIZE, "%lu", (unsigned long)value);
		return;
	}

	if (number == ObjectModuleParts) {
		if (objectAddress[MODULE_INDEX] >= CRATE_ADDESS_OFFSET) {
			bool isEch238 = false;
			CrateCache *crateCacheValue;

			if (objectAddress[ITEM_INDEX] >= DataCacheColumnCrate4xaSize) {
				strncpy(item->quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
				return;
			}

			objectAddress[MODULE_INDEX] -= CRATE_ADDESS_OFFSET;

			strncpy(item->quality, kdatacache_getCrateQuality(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
															  objectAddress[ITEM_INDEX], isEch238), QUALITY_SIZE);

			if (!strncmp(item->quality, ISEG_ITEM_QUALITY_INITIALIZE, QUALITY_SIZE)) {
				isEch238 = true;
				strncpy(item->quality, kdatacache_getCrateQuality(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
																  objectAddress[ITEM_INDEX], isEch238), QUALITY_SIZE);
			}
			crateCacheValue = kdatacache_crateCacheValue(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX], isEch238);

			DataType type = kdatacache_getCrateDataTypeEnum(objectAddress[ITEM_INDEX]);
			strncpy(item->timeStampLastRefreshed,
					kdatacache_getCrateTimeStampLastRefreshed(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
															  objectAddress[ITEM_INDEX], isEch238), TIME_SIZE);
			strncpy(item->timeStampLastChanged,
					kdatacache_getCrateTimeStampLastChanged(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
															objectAddress[ITEM_INDEX], isEch238), TIME_SIZE);

			if (type == DataTypeR4) {
				float value = kdatacache_getCrateValueR4(crateCacheValue, objectAddress[ITEM_INDEX], objectAddress[BYTE_ARRAY_INDEX]);
				snprintf(item->value, VALUE_SIZE, "%g", value);
				return;
			}

			if (type == DataTypeUI4 || type == DataTypeUI1 || type == DataTypeBool) {
				uint32_t value = kdatacache_getCrateValueUI4(crateCacheValue, objectAddress[ITEM_INDEX], objectAddress[BIT_ARRAY_INDEX]);
				snprintf(item->value, VALUE_SIZE, "%lu", (unsigned long)value);
				return;
			}

			if (type == DataTypeString) {
				strcpy(item->value, kdatacache_getCrateValueString(crateCacheValue, objectAddress[ITEM_INDEX]));
				return;
			}

			strcpy(item->quality, ISEG_ITEM_QUALITY_ERROR);
			return;
		}

		if (objectAddress[ITEM_INDEX] >= DataCacheColumnModuleSize) {
			strncpy(item->quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
			return;
		}

		DataType type = kdatacache_getModuleDataTypeEnum(objectAddress[ITEM_INDEX]);
		strncpy(item->timeStampLastRefreshed,
				kdatacache_getModuleTimeStampLastRefreshed(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
														   objectAddress[ITEM_INDEX]), TIME_SIZE);
		strncpy(item->timeStampLastChanged,
				kdatacache_getModuleTimeStampLastChanged(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
														 objectAddress[ITEM_INDEX]), TIME_SIZE);
		strncpy(item->quality, kdatacache_getModuleQuality(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
														   objectAddress[ITEM_INDEX]), QUALITY_SIZE);

		if (type == DataTypeR4) {
			float value = kdatacache_getModuleValueR4(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
													  objectAddress[ITEM_INDEX], objectAddress[BYTE_ARRAY_INDEX]);
			snprintf(item->value, VALUE_SIZE, "%g", value);
			return;
		}

		if (type == DataTypeUI4 || type == DataTypeBool) {
			uint32_t value = kdatacache_getModuleValueUI4(objectAddress[LINE_INDEX], objectAddress[MODULE_INDEX],
														  objectAddress[ITEM_INDEX], objectAddress[BIT_ARRAY_INDEX]);
			snprintf(item->value, VALUE_SIZE, "%lu", (unsigned long)value);
			return;
		}

		if (type == DataTypeString) {
			strcpy(item->value, kdatacache_getModuleValueString(objectAddress[LINE_INDEX],
																objectAddress[MODULE_INDEX], objectAddress[ITEM_INDEX]));
			return;
		}
	}

	if (number == ObjectLineParts) {
		if (objectAddress[ITEM_INDEX] >= DataCacheColumnLineSize) {
			strncpy(item->quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
			return;
		}

		if (objectAddress[ITEM_INDEX] == DataCacheColumnLineCanStatus) {
			strcpy(item->value, iseg_getLineProperty(kdatacache_getLinePropertyString(
														   objectAddress[ITEM_INDEX]), objectAddress[LINE_INDEX]));
			kdatacache_createLineTimeStamp(objectAddress[LINE_INDEX], objectAddress[ITEM_INDEX]);
			strncpy(item->timeStampLastRefreshed,
					kdatacache_getLineTimeStampLastRefreshed(objectAddress[LINE_INDEX],
															 objectAddress[ITEM_INDEX]), TIME_SIZE);
			strncpy(item->timeStampLastChanged,
					kdatacache_getSystemTimeStampLastChanged(DataCacheColumnSystemCanStatus), TIME_SIZE);
			strcpy(item->quality, ISEG_ITEM_QUALITY_OK);
			return;
		}
		if (objectAddress[ITEM_INDEX] == DataCacheColumnLineCanBitRate) {
			bool ok = false;
			snprintf(item->value, VALUE_SIZE, "%lu", (unsigned long)iseg_getBitRate(&ok));
			kdatacache_createLineTimeStamp(objectAddress[LINE_INDEX], objectAddress[ITEM_INDEX]);
			strncpy(item->timeStampLastRefreshed,
					kdatacache_getLineTimeStampLastRefreshed(objectAddress[LINE_INDEX],
															 objectAddress[ITEM_INDEX]), TIME_SIZE);
			strncpy(item->timeStampLastChanged,
					kdatacache_getSystemTimeStampLastChanged(DataCacheColumnSystemCanBitRate), TIME_SIZE);
			if (ok) {
				strcpy(item->quality, ISEG_ITEM_QUALITY_OK);
			} else {
				strcpy(item->quality, ISEG_ITEM_QUALITY_ERROR);
			}
			return;
		}

		DataType type = kdatacache_getLineDataTypeEnum(objectAddress[ITEM_INDEX]);
		strncpy(item->timeStampLastRefreshed,
				kdatacache_getLineTimeStampLastRefreshed(objectAddress[LINE_INDEX], objectAddress[ITEM_INDEX]), TIME_SIZE);
		strncpy(item->timeStampLastChanged,
				kdatacache_getLineTimeStampLastChanged(objectAddress[LINE_INDEX], objectAddress[ITEM_INDEX]), TIME_SIZE);
		strncpy(item->quality,
				kdatacache_getLineQuality(objectAddress[LINE_INDEX], objectAddress[ITEM_INDEX]), QUALITY_SIZE);

		if (type == DataTypeUI4) {
			snprintf(item->value, VALUE_SIZE, "%lu", (unsigned long)kdatacache_getDeviceValue(objectAddress[LINE_INDEX],
																							  objectAddress[ITEM_INDEX]));
			return;
		}

		if (type == DataTypeString) {
			strcpy(item->value, kdatacache_getLineValueString(objectAddress[LINE_INDEX], objectAddress[ITEM_INDEX],
																objectAddress[BIT_ARRAY_INDEX]));
			return;
		}
	}

	if (number == ObjectSystemParts) {

		if (objectAddress[ITEM_INDEX] >= DataCacheColumnSystemSize) {
			strncpy(item->quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
			return;
		}

		kdatacache_createSystemTimeStamp(false, objectAddress[ITEM_INDEX]);
		if (objectAddress[ITEM_INDEX] == DataCacheColumnSystemCanStatus) {
			strcpy(item->value, iseg_getLineProperty(kdatacache_getSystemPropertyString(objectAddress[ITEM_INDEX]), 0));	//TODO
			strncpy(item->timeStampLastRefreshed,
					kdatacache_getSystemTimeStampLastRefreshed(objectAddress[ITEM_INDEX]), TIME_SIZE);
			strncpy(item->timeStampLastChanged,
					kdatacache_getSystemTimeStampLastChanged(objectAddress[ITEM_INDEX]), TIME_SIZE);

			strcpy(item->quality, ISEG_ITEM_QUALITY_OK);
			return;
		}
		if (objectAddress[ITEM_INDEX] == DataCacheColumnSystemCanBitRate) {
			bool ok = false;
			snprintf(item->value, VALUE_SIZE, "%lu", (unsigned long)iseg_getBitRate(&ok));
			strncpy(item->timeStampLastRefreshed,
					kdatacache_getSystemTimeStampLastRefreshed(objectAddress[ITEM_INDEX]), TIME_SIZE);
			strncpy(item->timeStampLastChanged,
					kdatacache_getSystemTimeStampLastChanged(objectAddress[ITEM_INDEX]), TIME_SIZE);
			if (ok) {
				strcpy(item->quality, ISEG_ITEM_QUALITY_OK);
			} else {
				strcpy(item->quality, ISEG_ITEM_QUALITY_ERROR);
			}
			return;
		}

		DataType type = kdatacache_getSystemDataTypeEnum(objectAddress[ITEM_INDEX]);
		strncpy(item->timeStampLastChanged, kdatacache_getSystemTimeStampLastChanged(objectAddress[ITEM_INDEX]), TIME_SIZE);
		strncpy(item->timeStampLastRefreshed, kdatacache_getSystemTimeStampLastRefreshed(objectAddress[ITEM_INDEX]), TIME_SIZE);
		strncpy(item->quality, kdatacache_getSystemQuality(objectAddress[ITEM_INDEX]), QUALITY_SIZE);

		if (type == DataTypeUI4) {
			snprintf(item->value, VALUE_SIZE, "%lu", (unsigned long)kdatacache_getSystemValue(objectAddress[ITEM_INDEX]));
			return;
		}

		if (type == DataTypeString) {
			strcpy(item->value, kdatacache_getSystemValueString(objectAddress[ITEM_INDEX], objectAddress[BIT_ARRAY_INDEX]));
			return;
		}

	}

	strcpy(item->quality, ISEG_ITEM_QUALITY_ERROR);
	return;
}

/**
 * @brief read an item value, specified by the item object, from the data cache
 * @param item object
 * @return isegItemValue { object, value, quality, time }
 */
IsegItemValue iseg_getItemValue(const char *object)
{
	IsegItemValue item = EmptyIsegItemValue;
	uint16_t objectAddress[INDEX_SIZE] = { 0 };

	strncpy(item.object, object, FULLY_QUALIFIED_OBJECT_SIZE);
	int number = split(objectAddress, object, '.');

	if (objectAddress[LINE_INDEX] < m_interfaceIndex) {
		strncpy(item.quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
		return item;
	}

	objectAddress[LINE_INDEX] -= m_interfaceIndex;

	if (number == NoneObjectParts) {
		strcpy(item.quality, ISEG_ITEM_QUALITY_ERROR);
		return item;
	}

	kmutex_lock();
	itemValue(&item, objectAddress, number);
	kmutex_unlock();

	return item;
}

/**
 * @brief read an item property, specified by the item object, from the data cache
 * @param item object
 * @return isegItemProperty { object, type, unit, access }
 */
IsegItemProperty iseg_getItemProperty(const char *object)
{
	IsegItemProperty itemProperty = EmptyIsegItemProperty;
	uint16_t         objectAddress[INDEX_SIZE] = { 0 };

	strncpy(itemProperty.quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);

	strncpy(itemProperty.object, object, FULLY_QUALIFIED_OBJECT_SIZE);
	int number = split(objectAddress, object, '.');

	if (objectAddress[LINE_INDEX] < m_interfaceIndex)
		return itemProperty;

	objectAddress[LINE_INDEX] -= m_interfaceIndex;

	if (number < ObjectSystemParts || number > ObjectChannelParts) {
		return itemProperty;
	}

	if (number == ObjectChannelParts && objectAddress[ITEM_INDEX] < DataCacheColumnChannelSize) {
		strncpy(itemProperty.type, kdatacache_getChannelDataTypeString(objectAddress[ITEM_INDEX], objectAddress[BIT_ARRAY_INDEX]),
				DATA_TYPE_SIZE);
		strncpy(itemProperty.unit, kdatacache_getChannelUnitString(objectAddress[ITEM_INDEX]), UNIT_SIZE);
		strncpy(itemProperty.access, kdatacache_getChannelAccessString(objectAddress[ITEM_INDEX]), ACCESS_SIZE);
		strncpy(itemProperty.quality, ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
		return itemProperty;
	}

	if (number == ObjectModuleParts) {
		if (objectAddress[MODULE_INDEX] >= CRATE_ADDESS_OFFSET) {
			objectAddress[MODULE_INDEX] -= CRATE_ADDESS_OFFSET;
			if (objectAddress[ITEM_INDEX] < DataCacheColumnCrate4xaSize) {
				strncpy(itemProperty.type, kdatacache_getCrateDataTypeString(objectAddress[ITEM_INDEX], objectAddress[BIT_ARRAY_INDEX]),
						DATA_TYPE_SIZE);
				strncpy(itemProperty.unit, kdatacache_getCrateUnitString(objectAddress[ITEM_INDEX]), UNIT_SIZE);
				strncpy(itemProperty.access, kdatacache_getCrateAccessString(objectAddress[ITEM_INDEX]), ACCESS_SIZE);
				strncpy(itemProperty.quality, ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
			}

			return itemProperty;
		}

		if (objectAddress[ITEM_INDEX] < DataCacheColumnModuleSize) {
			strncpy(itemProperty.type, kdatacache_getModuleDataTypeString(objectAddress[ITEM_INDEX], objectAddress[BIT_ARRAY_INDEX])
					, DATA_TYPE_SIZE);
			strncpy(itemProperty.unit, kdatacache_getModuleUnitString(objectAddress[ITEM_INDEX]), UNIT_SIZE);
			strncpy(itemProperty.access, kdatacache_getModuleAccessString(objectAddress[ITEM_INDEX]), ACCESS_SIZE);
			strncpy(itemProperty.quality, ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
			return itemProperty;
		}
	}

	if (number == ObjectLineParts) {
		strncpy(itemProperty.type,
			kdatacache_getLineDataTypeString(objectAddress[ITEM_INDEX]), DATA_TYPE_SIZE);
		strncpy(itemProperty.unit,
			kdatacache_getLineUnitString(objectAddress[ITEM_INDEX]), UNIT_SIZE);
		strncpy(itemProperty.access,
			kdatacache_getLineAccessString(objectAddress[ITEM_INDEX]), ACCESS_SIZE);
		strncpy(itemProperty.quality, ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
		return itemProperty;
	}

	if (number == ObjectSystemParts) {
		strncpy(itemProperty.type,
			kdatacache_getSystemDataTypeString(objectAddress[ITEM_INDEX]), DATA_TYPE_SIZE);
		strncpy(itemProperty.unit,
			kdatacache_getSystemUnitString(objectAddress[ITEM_INDEX]), UNIT_SIZE);
		strncpy(itemProperty.access,
			kdatacache_getSystemAccessString(objectAddress[ITEM_INDEX]), ACCESS_SIZE);
		strncpy(itemProperty.quality, ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
		return itemProperty;
	}

	return itemProperty;
}
