#include "isegclientapi.h"
#include "version.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <signal.h>
#include <pthread.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define BUF 512
int  m_socketDescriptor = 0;
static pthread_mutex_t m_mutexSocket = PTHREAD_MUTEX_INITIALIZER;

/**
 * @brief Returns the library four digit version number as unsigned int value.
 *
 * This allows to check for a specific version number.
 * @since 1.1
 * @return The version number as unsigned int value, e.g. 0x01010000
 */
unsigned int iseg_getVersion(void)
{
	const unsigned int version[] = { VER_FILEVERSION };

	return (version[0] << 24) + (version[1] << 16) + (version[2] << 8) + version[3];
}

/**
 * @brief Returns the library four digit version number as null-terminated string.
 *
 * The digits are separated by dots.
 * @since 1.1
 * @return The version number, e.g. "1.1.0.3"
 */
const char *iseg_getVersionString(void)
{
	return VER_FILEVERSION_STR;
}

int iseg_connect(const char* dummy)
{
	(void)dummy;
	struct sockaddr_un address;

	address.sun_family = AF_LOCAL;
	strcpy(address.sun_path, UDS_FILE);
	if ((m_socketDescriptor = socket(PF_LOCAL, SOCK_STREAM, 0)) < 0) {
		//printf("Fehler bei socket()");// ...(%s)\n", strerror(errno));
		return ISEG_ERROR;
	}

	if (connect(m_socketDescriptor, (struct sockaddr *) &address, sizeof(address))) {
		//printf("Fehler bei connect");// bei connect() ...(%s)\n", strerror(errno));
		return ISEG_ERROR;
	}
	return ISEG_OK;
}

int iseg_disconnect(void)
{
	close(m_socketDescriptor);
	return ISEG_OK;
}

/**
 * @brief writes an item value, specified by the item object, into the data cache
 * @param item - item struct
 * @return - 0 = ok
 *         - 1 = error
 */
int iseg_setItemValue(const IsegItemValue *item)
{
	char buf[BUF] = "s";
	char recvBuf[3];
	pthread_mutex_lock(&m_mutexSocket);
	memcpy(buf + 1, item, sizeof(IsegItemValue));
	send(m_socketDescriptor, buf, sizeof(IsegItemValue) + 1, 0);
	recv(m_socketDescriptor, recvBuf, 3, 0);
	//if (strcmp(recvBuf, "0") != 0) {
	//	printf("Error client API %s\n", recvBuf);
	//}
	pthread_mutex_unlock(&m_mutexSocket);
	return atoi(recvBuf);
}

/**
 * @brief read an item value, specified by the item object, from the data cache
 * @param item object
 * @return isegItemValue { object, value, quality, time }
 */
IsegItemValue iseg_getItemValue(const char *object)
{
	IsegItemValue item = EmptyIsegItemValue;
	char buf[BUF] = "g";
	char recvBuf[BUF];
	pthread_mutex_lock(&m_mutexSocket);
	memcpy(buf+1, object, strlen(object));
	send(m_socketDescriptor, buf, strlen(buf), 0);
	recv(m_socketDescriptor, recvBuf, sizeof(IsegItemValue), 0);
	memcpy(&item, recvBuf, sizeof(IsegItemValue));
	pthread_mutex_unlock(&m_mutexSocket);
	return item;
}

/**
 * @brief read an item property, specified by the item object, from the data cache
 * @param item object
 * @return isegItemProperty { object, type, unit, access }
 */
IsegItemProperty iseg_getItemProperty(const char *object)
{
	IsegItemProperty itemProperty = EmptyIsegItemProperty;
	char buf[BUF] = "p";
	char recvBuf[BUF];
	pthread_mutex_lock(&m_mutexSocket);
	memcpy(buf+1, object, strlen(object));
	//printf("Client send Prop %s\n", buf);
	send(m_socketDescriptor, buf, strlen(buf), 0);
	recv(m_socketDescriptor, recvBuf, sizeof(IsegItemProperty), 0);
	memcpy(&itemProperty, recvBuf, sizeof(IsegItemProperty));
	pthread_mutex_unlock(&m_mutexSocket);
	return itemProperty;
}
