#include "receivethread.h"
#include "isegcommon.h"
#include "kdatacache.h"

#include <stdio.h>
#include <string.h>

#include <linux/can/raw.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>

static int m_socket;
static struct msghdr m_msg;
static struct iovec m_iov;
static struct can_frame m_frame;
static struct sockaddr_can m_addr;
static char m_ctrlmsg[CMSG_SPACE(sizeof(struct timeval)) + CMSG_SPACE(sizeof(__u32))];

void writeLogMessage(const char *messageType, uint32_t level, const struct can_frame *frame, int count, uint32_t answers);

int receivethread_init(const char *canInterface)
{
	if ((m_socket = socket(PF_CAN, SOCK_RAW, CAN_RAW)) < 0) {
		fprintf(stderr, "receive: error socket()\n");
		return 0;
	}

	struct ifreq ifr;
	memset(&ifr.ifr_name, 0, sizeof(ifr.ifr_name));
	strcpy(ifr.ifr_name, canInterface);

	if (ioctl(m_socket, SIOCGIFINDEX, &ifr) < 0) {
		fprintf(stderr, "receive: error SIOCGIFINDEX\n");
		return 0;
	}

	m_addr.can_family = AF_CAN;
	m_addr.can_ifindex = ifr.ifr_ifindex;

	if (bind(m_socket, (struct sockaddr *)&m_addr, sizeof(m_addr)) < 0) {
		fprintf(stderr, "receive: error bind\n");
		return 0;
	}

	m_iov.iov_base = &m_frame;
	m_msg.msg_name = &m_addr;
	m_msg.msg_iov = &m_iov;
	m_msg.msg_iovlen = 1;
	m_msg.msg_control = &m_ctrlmsg;

	return 1;
}

static KCanMessage toKCanMessage(struct can_frame frame)
{
	enum { BitDcp = 0x80 };

	KCanMessage result = EmptyCanMessage;

	const int size = frame.can_dlc;
	result.id.w = frame.can_id;
	result.bit.dlc = frame.can_dlc;

	if (size >= 1 && (frame.data[0] & BitDcp)) {
		result.bit.edcp = false;
		result.dataId = 0x8000 + (frame.data[0] & 0xFF);
		memcpy(result.data, &frame.data[1], size - 1);
	} else if (size >= 2 && !(frame.data[0] & BitDcp)) {
		result.bit.edcp = true;
		result.dataId = (frame.data[0] << 8) + frame.data[1];
		memcpy(result.data, &frame.data[2], size - 2);
	}

	result.bit.busy = true;
	return result;
}

KCanMessage receivethread_waitForMessage(void)
{
	const KCanMessage emptyResult = EmptyCanMessage;
	fd_set rdfs;
	static int count = 0;

	FD_ZERO(&rdfs);
	FD_SET(m_socket, &rdfs);

	memset(&m_frame, 0, sizeof(m_frame));

	struct timeval timeout = { 1, 0000 };
	int ret;
	if ((ret = select(m_socket + 1, &rdfs, NULL, NULL, &timeout)) < 0) {
		fprintf(stderr, "receive: error select\n");
		return emptyResult;
	}

	if (!FD_ISSET(m_socket, &rdfs))
		return emptyResult;

	/* these settings may be modified by recvmsg() */
	m_iov.iov_len = sizeof(m_frame);
	m_msg.msg_namelen = sizeof(m_addr);
	m_msg.msg_controllen = sizeof(m_ctrlmsg);
	m_msg.msg_flags = 0;

	int nbytes = recvmsg(m_socket, &m_msg, 0);
	if (nbytes < 0) {
		fprintf(stderr, "receive: error nothing to read\n");
		return emptyResult;
	}

	if (nbytes < (int)sizeof(struct can_frame)) {
		fprintf(stderr, "receive: error incomplete CAN frame\n");
		return emptyResult;
	}

	uint32_t logLevel = kdatacache_getSystemValue(DataCacheColumnSystemLogLevel);
	uint32_t _answers = kcache_remainingAnswers(m_frame.can_id & 1024);
	if (logLevel & LogCanReceive && (m_frame.can_id & 4) ) {
		writeLogMessage("CANRX", logLevel, &m_frame, ++count, _answers);
	}

	return toKCanMessage(m_frame);
}
