#include "sendthread.h"

#include <kcan.h>
#include<isegcommon.h>
#include <kdatacache.h>

#include <stdio.h>
#include <string.h>

#include <linux/can.h>
#include <linux/can/raw.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <unistd.h>

enum { Debug = 0 };

static int m_socket; /* can raw socket */
static struct sockaddr_can addr;
static struct ifreq ifr;

void writeLogMessage(const char *messageType, uint32_t level, const struct can_frame *frame, int count, uint32_t answers);

int sendthread_init(const char *canInterface)
{
	/* open socket */
	if ((m_socket = socket(PF_CAN, SOCK_RAW, CAN_RAW)) < 0) {
		fprintf(stderr, "send: error opening socket\n");
		return 0;
	}

	addr.can_family = AF_CAN;

	strcpy(ifr.ifr_name, canInterface);
	if (ioctl(m_socket, SIOCGIFINDEX, &ifr) < 0) {
		fprintf(stderr, "send: error SIOCGIFINDEX\n");
		return 0;
	}
	addr.can_ifindex = ifr.ifr_ifindex;

	/* disable default receive filter on this RAW socket */
	/* This is obsolete as we do not read from the socket at all, but for */
	/* this reason we can remove the receive list in the Kernel to save a */
	/* little (really a very little!) CPU usage.                          */
	setsockopt(m_socket, SOL_CAN_RAW, CAN_RAW_FILTER, NULL, 0);

	// disable loopback, otherwise all frames immediately occur in the receive thread
	int loopback = 0;
	setsockopt(m_socket, SOL_CAN_RAW, CAN_RAW_LOOPBACK, &loopback, sizeof(loopback));

	if (bind(m_socket, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		fprintf(stderr, "send: error bind\n");
		return 0;
	}

	return 1;
}

static struct can_frame toCanFrame(KCanMessage *msg)
{
	const int size = msg->bit.dlc;

	struct can_frame frame = { 0 };
	frame.can_id = msg->id.w;
	frame.can_id |= msg->bit.rtr << 30;
	frame.can_id |= msg->bit.extendedId << 31;
	frame.can_dlc = size;
	if (msg->bit.edcp && size >= 2) {
		frame.data[0] = msg->dataId >> 8;
		frame.data[1] = msg->dataId & 0xFF;
		memcpy(&frame.data[2], msg->data, size - 2);
	} else if (!msg->bit.edcp && size >= 1) {
		frame.data[0] = msg->dataId & 0xFF;
		memcpy(&frame.data[1], msg->data, size - 1);
	}

	return frame;
}

void sendthread_send(KCanMessage *msg)
{
	static bool communicationLost = false;
	static int counterTx = 0;
	static int counterRq = 0;
	struct can_frame frame = toCanFrame(msg);

	uint32_t logLevel = kdatacache_getSystemValue(DataCacheColumnSystemLogLevel);
	uint32_t answers = kcache_remainingAnswers(msg->id.bit.crate);
	if (logLevel & LogCanWriteRequest && ((frame.can_id & 1) == 0) ) {
		writeLogMessage("CANTX", logLevel, &frame, ++counterTx, answers);
	} else if (logLevel & LogCanReadRequest && (frame.can_id & 1) ) {
		writeLogMessage("CANRQ", logLevel, &frame, ++counterRq, answers);
	}

	int nbytes;
	if ((nbytes = write(m_socket, &frame, sizeof(frame))) != sizeof(frame)) {
		if (!communicationLost) {
			fprintf(stderr, "send: write error!");
			kdatacache_createSystemTimeStamp(true, DataCacheColumnSystemCanStatus);
			communicationLost = true;
		}
		return;
	}

	if (communicationLost)
		kdatacache_createSystemTimeStamp(true, DataCacheColumnSystemCanStatus);

	communicationLost = false;
}
